@@ -1,6 +1,6 @@
 /*
  * pidstat: Report statistics for Linux tasks
- * (C) 2007-2017 by Sebastien GODARD (sysstat <at> orange.fr)
+ * (C) 2007-2018 by Sebastien GODARD (sysstat <at> orange.fr)
  *
  ***************************************************************************
  * This program is free software; you can redistribute it and/or modify it *
@@ -52,8 +52,8 @@
 char *sccsid(void) { return (SCCSID); }
 #endif
 
-unsigned long long uptime[3] = {0, 0, 0};
-unsigned long long uptime0[3] = {0, 0, 0};
+unsigned long long tot_jiffies[3] = {0, 0, 0};
+unsigned long long uptime_cs[3] = {0, 0, 0};
 struct pid_stats *st_pid_list[3] = {NULL, NULL, NULL};
 unsigned int *pid_array = NULL;
 struct pid_stats st_pid_null;
@@ -76,6 +76,8 @@ unsigned int actflag = 0;	/* Activity flag */
 struct sigaction alrm_act, int_act, chld_act;
 int signal_caught = 0;
 
+int dplaces_nr = -1;		/* Number of decimal places */
+
 /*
  ***************************************************************************
  * Print usage and exit.
@@ -91,14 +93,15 @@ void usage(char *progname)
 
 	fprintf(stderr, _("Options are:\n"
 			  "[ -d ] [ -H ] [ -h ] [ -I ] [ -l ] [ -R ] [ -r ] [ -s ] [ -t ] [ -U [ <username> ] ]\n"
-			  "[ -u ] [ -V ] [ -v ] [ -w ] [ -C <command> ] [ -G <process_name> ] [ --human ]\n"
-			  "[ -p { <pid> [,...] | SELF | ALL } ] [ -T { TASK | CHILD | ALL } ]\n"));
+			  "[ -u ] [ -V ] [ -v ] [ -w ] [ -C <command> ] [ -G <process_name> ]\n"
+			  "[ -p { <pid> [,...] | SELF | ALL } ] [ -T { TASK | CHILD | ALL } ]\n"
+			  "[ --dec={ 0 | 1 | 2 } ] [ --human ]\n"));
 	exit(1);
 }
 
 /*
  ***************************************************************************
- * SIGALRM signal handler. No need to resert the handler here.
+ * SIGALRM signal handler. No need to reset the handler here.
  *
  * IN:
  * @sig	Signal number.
@@ -1009,8 +1012,18 @@ void read_stats(int curr)
 		perror("malloc");
 		exit(4);
 	}
-	/* Read statistics for CPUs "all" and 0 */
-	read_stat_cpu(st_cpu, 2, &uptime[curr], &uptime0[curr]);
+	/* Read statistics for CPUs "all" */
+	read_stat_cpu(st_cpu, 1);
+
+	/*
+	 * Compute the total number of jiffies spent by all processors.
+	 * NB: Don't add cpu_guest/cpu_guest_nice because cpu_user/cpu_nice
+	 * already include them.
+	 */
+	tot_jiffies[curr] = st_cpu->cpu_user + st_cpu->cpu_nice +
+			    st_cpu->cpu_sys + st_cpu->cpu_idle +
+			    st_cpu->cpu_iowait + st_cpu->cpu_hardirq +
+			    st_cpu->cpu_steal + st_cpu->cpu_softirq;
 	free(st_cpu);
 
 	if (DISPLAY_ALL_PID(pidflag)) {
@@ -1391,8 +1404,8 @@ void print_line_id(char *timestamp, struct pid_stats *pst)
  * @curr_string	String displayed at the beginning of current sample stats.
  * 		This is the timestamp of the current sample.
  * @itv		Interval of time in jiffies.
- * @g_itv	Interval of time in jiffies multiplied by the number of
- * 		processors.
+ * @deltot_jiffies
+ *		Number of jiffies spent on the interval by all processors.
  *
  * RETURNS:
  * 0 if all the processes to display have terminated.
@@ -1402,7 +1415,7 @@ void print_line_id(char *timestamp, struct pid_stats *pst)
 int write_pid_task_all_stats(int prev, int curr, int dis,
 			     char *prev_string, char *curr_string,
 			     unsigned long long itv,
-			     unsigned long long g_itv)
+			     unsigned long long deltot_jiffies)
 {
 	struct pid_stats *pstc, *pstp;
 	char dstr[32];
@@ -1448,16 +1461,16 @@ int write_pid_task_all_stats(int prev, int curr, int dis,
 				   (pstc->utime - pstc->gtime) < (pstp->utime - pstp->gtime) ?
 				   0.0 :
 				   SP_VALUE(pstp->utime - pstp->gtime,
-					    pstc->utime - pstc->gtime, itv),
-				   SP_VALUE(pstp->stime,  pstc->stime, itv),
-				   SP_VALUE(pstp->gtime,  pstc->gtime, itv),
-				   SP_VALUE(pstp->wtime, pstc->wtime, itv),
+					    pstc->utime - pstc->gtime, itv * HZ / 100),
+				   SP_VALUE(pstp->stime,  pstc->stime, itv * HZ / 100),
+				   SP_VALUE(pstp->gtime,  pstc->gtime, itv * HZ / 100),
+				   SP_VALUE(pstp->wtime, pstc->wtime, itv * HZ / 100),
 				   /* User time already includes guest time */
 				   IRIX_MODE_OFF(pidflag) ?
 				   SP_VALUE(pstp->utime + pstp->stime,
-					    pstc->utime + pstc->stime, g_itv) :
+					    pstc->utime + pstc->stime, deltot_jiffies) :
 				   SP_VALUE(pstp->utime + pstp->stime,
-					    pstc->utime + pstc->stime, itv));
+					    pstc->utime + pstc->stime, itv * HZ / 100));
 
 			cprintf_in(IS_INT, "   %3d", "", pstc->processor);
 		}
@@ -1625,9 +1638,9 @@ int write_pid_child_all_stats(int prev, int curr, int dis,
  * @curr_string	String displayed at the beginning of current sample stats.
  * 		This is the timestamp of the current sample, or "Average"
  * 		when displaying average stats.
- * @itv		Interval of time in jiffies.
- * @g_itv	Interval of time in jiffies multiplied by the number of
- * 		processors.
+ * @itv		Interval of time in 1/100th of a second.
+ * @deltot_jiffies
+ *		Number of jiffies spent on the interval by all processors.
  *
  * RETURNS:
  * 0 if all the processes to display have terminated.
@@ -1637,7 +1650,7 @@ int write_pid_child_all_stats(int prev, int curr, int dis,
 int write_pid_task_cpu_stats(int prev, int curr, int dis, int disp_avg,
 			     char *prev_string, char *curr_string,
 			     unsigned long long itv,
-			     unsigned long long g_itv)
+			     unsigned long long deltot_jiffies)
 {
 	struct pid_stats *pstc, *pstp;
 	unsigned int p;
@@ -1659,16 +1672,16 @@ int write_pid_task_cpu_stats(int prev, int curr, int dis, int disp_avg,
 			   (pstc->utime - pstc->gtime) < (pstp->utime - pstp->gtime) ?
 			   0.0 :
 			   SP_VALUE(pstp->utime - pstp->gtime,
-				    pstc->utime - pstc->gtime, itv),
-			   SP_VALUE(pstp->stime, pstc->stime, itv),
-			   SP_VALUE(pstp->gtime, pstc->gtime, itv),
-			   SP_VALUE(pstp->wtime, pstc->wtime, itv),
+				    pstc->utime - pstc->gtime, itv * HZ / 100),
+			   SP_VALUE(pstp->stime, pstc->stime, itv * HZ / 100),
+			   SP_VALUE(pstp->gtime, pstc->gtime, itv * HZ / 100),
+			   SP_VALUE(pstp->wtime, pstc->wtime, itv * HZ / 100),
 			   /* User time already includes guest time */
 			   IRIX_MODE_OFF(pidflag) ?
 			   SP_VALUE(pstp->utime + pstp->stime,
-				    pstc->utime + pstc->stime, g_itv) :
+				    pstc->utime + pstc->stime, deltot_jiffies) :
 			   SP_VALUE(pstp->utime + pstp->stime,
-				    pstc->utime + pstc->stime, itv));
+				    pstc->utime + pstc->stime, itv * HZ / 100));
 
 		if (!disp_avg) {
 			cprintf_in(IS_INT, "   %3d", "", pstc->processor);
@@ -1783,7 +1796,7 @@ int write_pid_child_cpu_stats(int prev, int curr, int dis, int disp_avg,
  * @curr_string	String displayed at the beginning of current sample stats.
  * 		This is the timestamp of the current sample, or "Average"
  * 		when displaying average stats.
- * @itv		Interval of time in jiffies.
+ * @itv		Interval of time in 1/100th of a second.
  *
  * RETURNS:
  * 0 if all the processes to display have terminated.
@@ -2008,7 +2021,7 @@ int write_pid_stack_stats(int prev, int curr, int dis, int disp_avg,
  * @curr_string	String displayed at the beginning of current sample stats.
  * 		This is the timestamp of the current sample, or "Average"
  * 		when displaying average stats.
- * @itv		Interval of time in jiffies.
+ * @itv		Interval of time in 1/100th of a second.
  *
  * RETURNS:
  * 0 if all the processes to display have terminated.
@@ -2097,7 +2110,7 @@ int write_pid_io_stats(int prev, int curr, int dis, int disp_avg,
  * @curr_string	String displayed at the beginning of current sample stats.
  * 		This is the timestamp of the current sample, or "Average"
  * 		when displaying average stats.
- * @itv		Interval of time in jiffies.
+ * @itv		Interval of time in 1/100th of a second.
  *
  * RETURNS:
  * 0 if all the processes to display have terminated.
@@ -2287,23 +2300,16 @@ int write_pid_ktab_stats(int prev, int curr, int dis, int disp_avg,
 int write_stats_core(int prev, int curr, int dis, int disp_avg,
 		     char *prev_string, char *curr_string)
 {
-	unsigned long long itv, g_itv;
+	unsigned long long itv, deltot_jiffies;
 	int again = 0;
 
 	/* Test stdout */
 	TEST_STDOUT(STDOUT_FILENO);
 
-	/* g_itv is multiplied by the number of processors */
-	g_itv = get_interval(uptime[prev], uptime[curr]);
+	/* Total number of jiffies spent on the interval */
+	deltot_jiffies = get_interval(tot_jiffies[prev], tot_jiffies[curr]);
 
-	if (cpu_nr > 1) {
-		/* SMP machines */
-		itv = get_interval(uptime0[prev], uptime0[curr]);
-	}
-	else {
-		/* UP machines */
-		itv = g_itv;
-	}
+	itv = get_interval(uptime_cs[prev], uptime_cs[curr]);
 
 	if (PROCESS_STRING(pidflag)) {
 		/* Reset "show threads" flag */
@@ -2313,7 +2319,7 @@ int write_stats_core(int prev, int curr, int dis, int disp_avg,
 	if (DISPLAY_ONELINE(pidflag)) {
 		if (DISPLAY_TASK_STATS(tskflag)) {
 			again += write_pid_task_all_stats(prev, curr, dis, prev_string, curr_string,
-							  itv, g_itv);
+							  itv, deltot_jiffies);
 		}
 		if (DISPLAY_CHILD_STATS(tskflag)) {
 			again += write_pid_child_all_stats(prev, curr, dis, prev_string, curr_string);
@@ -2326,7 +2332,7 @@ int write_stats_core(int prev, int curr, int dis, int disp_avg,
 			if (DISPLAY_TASK_STATS(tskflag)) {
 				again += write_pid_task_cpu_stats(prev, curr, dis, disp_avg,
 								  prev_string, curr_string,
-								  itv, g_itv);
+								  itv, deltot_jiffies);
 			}
 			if (DISPLAY_CHILD_STATS(tskflag)) {
 				again += write_pid_child_cpu_stats(prev, curr, dis, disp_avg,
@@ -2373,8 +2379,7 @@ int write_stats_core(int prev, int curr, int dis, int disp_avg,
 
 		/* Display scheduling priority and policy information */
 		if (DISPLAY_RT(actflag)) {
-			again += write_pid_rt_stats(prev, curr, dis, prev_string,
-						    curr_string);
+			again += write_pid_rt_stats(prev, curr, dis, prev_string, curr_string);
 		}
 	}
 
@@ -2469,15 +2474,8 @@ void rw_pidstat_loop(int dis_hdr, int rows)
 	/* Don't buffer data if redirected to a pipe */
 	setbuf(stdout, NULL);
 
-	if (cpu_nr > 1) {
-		/*
-		 * Read system uptime (only for SMP machines).
-		 * Init uptime0. So if /proc/uptime cannot fill it, this will be
-		 * done by /proc/stat.
-		 */
-		uptime0[0] = 0;
-		read_uptime(&uptime0[0]);
-	}
+	/* Read system uptime */
+	read_uptime(&uptime_cs[0]);
 	read_stats(0);
 
 	if (DISPLAY_MEM(actflag)) {
@@ -2501,8 +2499,8 @@ void rw_pidstat_loop(int dis_hdr, int rows)
 
 	/* Save the first stats collected. Will be used to compute the average */
 	ps_tstamp[2] = ps_tstamp[0];
-	uptime[2] = uptime[0];
-	uptime0[2] = uptime0[0];
+	tot_jiffies[2] = tot_jiffies[0];
+	uptime_cs[2] = uptime_cs[0];
 	memcpy(st_pid_list[2], st_pid_list[0], PID_STATS_SIZE * pid_nr);
 
 	/* Set a handler for SIGINT */
@@ -2521,15 +2519,8 @@ void rw_pidstat_loop(int dis_hdr, int rows)
 		/* Get time */
 		get_localtime(&ps_tstamp[curr], 0);
 
-		if (cpu_nr > 1) {
-			/*
-			 * Read system uptime (only for SMP machines).
-			 * Init uptime0. So if /proc/uptime cannot fill it, this will be
-			 * done by /proc/stat.
-			 */
-			uptime0[curr] = 0;
-			read_uptime(&(uptime0[curr]));
-		}
+		/* Read system uptime (in 1/100th of a second) */
+		read_uptime(&(uptime_cs[curr]));
 
 		/* Read stats */
 		read_stats(curr);
@@ -2735,6 +2726,15 @@ int main(int argc, char **argv)
 			opt++;
 		}
 
+		else if (!strncmp(argv[opt], "--dec=", 6) && (strlen(argv[opt]) == 7)) {
+			/* Get number of decimal places */
+			dplaces_nr = atoi(argv[opt] + 6);
+			if ((dplaces_nr < 0) || (dplaces_nr > 2)) {
+				usage(argv[0]);
+			}
+			opt++;
+		}
+
 		else if (!strcmp(argv[opt], "-T")) {
 			if (!argv[++opt]) {
 				usage(argv[0]);
